import os

from zrb.helper.codemod.add_key_value_to_dict import add_key_value_to_dict
from zrb.helper.file.text import read_text_file_async, write_text_file_async
from zrb.helper.typecheck import typechecked
from zrb.helper.util import to_snake_case
from zrb.task.task import Task

from ._common import get_app_dir, get_python_value_for_testing


@typechecked
async def inject_test(
    task: Task,
    project_dir: str,
    app_name: str,
    module_name: str,
    entity_name: str,
    column_name: str,
    column_type: str,
):
    snake_module_name = to_snake_case(module_name)
    snake_entity_name = to_snake_case(entity_name)
    snake_column_name = to_snake_case(column_name)
    test_file_path = os.path.join(
        get_app_dir(project_dir, app_name),
        "test",
        snake_module_name,
        f"test_{snake_entity_name}.py",
    )
    task.print_out(f"Read code from: {test_file_path}")
    code = await read_text_file_async(test_file_path)
    task.print_out(f'Add column "{snake_column_name}" to the test')
    dict_names = [
        "inserted_success_data",
        "to_be_updated_success_data",
        "updated_success_data",
        "to_be_deleted_success_data",
    ]
    default_python_value = get_python_value_for_testing(column_type)
    for dict_name in dict_names:
        code = add_key_value_to_dict(
            code=code,
            dict_name=dict_name,
            key=f"'{snake_column_name}'",
            value=default_python_value,
        )
    task.print_out(f"Write modified code to: {test_file_path}")
    await write_text_file_async(test_file_path, code)



from component.model import RepoModel
from module.log.schema.activity import Activity, ActivityData, ActivityResult


class ActivityModel(RepoModel[Activity, ActivityData, ActivityResult]):
    schema_result_cls = ActivityResult



import jsons

from zrb import Env, EnvFile

from ._constant import (
    APP_TEMPLATE_ENV_FILE_NAME,
    DEPLOYMENT_DIR,
    DEPLOYMENT_TEMPLATE_ENV_FILE_NAME,
    MODULES,
)

app_env_file = EnvFile(path=APP_TEMPLATE_ENV_FILE_NAME, prefix="ZRB_ENV_PREFIX")

app_enable_otel_env = Env(
    name="APP_ENABLE_OTEL",
    default='{{ "1" if input.enable_snake_zrb_app_name_monitoring else "0" }}',
)

deployment_app_env_file = EnvFile(
    path=APP_TEMPLATE_ENV_FILE_NAME, prefix="DEPLOYMENT_APP_ZRB_ENV_PREFIX"
)

deployment_config_env_file = EnvFile(
    path=DEPLOYMENT_TEMPLATE_ENV_FILE_NAME, prefix="DEPLOYMENT_CONFIG_ZRB_ENV_PREFIX"
)

deployment_enable_monitoring_env = Env(
    name="ENABLE_MONITORING",
    os_name="DEPLOYMENT_CONFIG_ZRB_ENV_PREFIX_ENABLE_MONITORING",
    default="{{ 1 if input.enable_snake_zrb_app_name_monitoring else 0 }}",
)

pulumi_backend_url_env = Env(
    name="PULUMI_BACKEND_URL",
    os_name="PULUMI_ZRB_ENV_PREFIX_BACKEND_URL",
    default=f"file://{DEPLOYMENT_DIR}/state",
)

pulumi_config_passphrase_env = Env(
    name="PULUMI_CONFIG_PASSPHRASE",
    os_name="PULUMI_ZRB_ENV_PREFIX_CONFIG_PASSPHRASE",
    default="secret",
)

deployment_modules_env = Env(
    name="MODULES",
    os_name="DEPLOYMENT_CONFIG_ZRB_ENV_PREFIX_MODULES",
    default=jsons.dumps(MODULES),
)

deployment_modules_env = Env(
    name="MODULES",
    os_name="DEPLOYMENT_CONFIG_ZRB_ENV_PREFIX_MODULES",
    default=jsons.dumps(MODULES),
)



from ._group import snake_zrb_app_name_frontend_group
from .build import build_snake_zrb_app_name_frontend
from .build_once import build_snake_zrb_app_name_frontend_once

assert snake_zrb_app_name_frontend_group
assert build_snake_zrb_app_name_frontend
assert build_snake_zrb_app_name_frontend_once



import asyncio
import atexit
import logging
import os
import pathlib
import signal
import subprocess
import sys
import time
from collections.abc import Callable, Iterable
from typing import Any, Optional, TypeVar, Union

from zrb.config.config import DEFAULT_SHELL, LOGGING_LEVEL
from zrb.helper.accessories.color import colored
from zrb.helper.asyncio_task import stop_asyncio_sync
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked
from zrb.helper.typing import JinjaTemplate
from zrb.task.any_task import AnyTask
from zrb.task.any_task_event_handler import (
    OnFailed,
    OnReady,
    OnRetry,
    OnSkipped,
    OnStarted,
    OnTriggered,
    OnWaiting,
)
from zrb.task.base_task.base_task import BaseTask
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_group.group import Group
from zrb.task_input.any_input import AnyInput

logger.debug(colored("Loading zrb.task.cmd_task", attrs=["dark"]))

_has_stty = True
try:
    _original_stty = subprocess.getoutput("stty -g").rstrip()
except Exception:
    _has_stty = False


def _reset_stty():
    global _has_stty
    if _has_stty:
        try:
            subprocess.run(["stty", _original_stty], check=True)
        except Exception:
            _has_stty = False


def _log_error(message: Any):
    if LOGGING_LEVEL > logging.ERROR:
        return
    colored_message = colored(f"{message}", color="red", attrs=["bold"])
    logger.error(colored_message, exc_info=True)


def _print_out_dark(message: Any):
    message_str = f"{message}"
    print(colored(message_str, attrs=["dark"]), file=sys.stderr)


CmdVal = Union[
    JinjaTemplate,
    Iterable[JinjaTemplate],
    Callable[..., Union[Iterable[JinjaTemplate], JinjaTemplate]],
]
TCmdTask = TypeVar("TCmdTask", bound="CmdTask")


class CmdResult:
    def __init__(self, output: str, error: str):
        self.output = output
        self.error = error

    def __str__(self) -> str:
        return self.output


class CmdGlobalState:
    def __init__(self):
        self.no_more_attempt: bool = False
        self.is_killed_by_signal: bool = False


@typechecked
class CmdTask(BaseTask):
    """
    Command Task.
    You can use this task to run shell command.

    Examples:
        >>> from zrb import runner, CmdTask, StrInput, Env
        >>> hello = CmdTask(
        >>>     name='hello',
        >>>     inputs=[StrInput(name='name', default='World')],
        >>>     envs=[Env(name='HOME_DIR', os_name='HOME')],
        >>>     cmd=[
        >>>         'echo Hello {{ input.name }}',
        >>>         'echo Home directory is: $HOME_DIR',
        >>>     ]
        >>> )
        >>> runner.register(hello)
    """

    _pids: list[int] = []
    _global_state = CmdGlobalState()

    def __init__(
        self,
        name: str,
        group: Optional[Group] = None,
        inputs: Iterable[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = "",
        executable: Optional[str] = None,
        cmd: CmdVal = "",
        cmd_path: CmdVal = "",
        cwd: Optional[Union[JinjaTemplate, pathlib.Path]] = None,
        should_render_cwd: bool = True,
        upstreams: Iterable[AnyTask] = [],
        fallbacks: Iterable[AnyTask] = [],
        on_triggered: Optional[OnTriggered] = None,
        on_waiting: Optional[OnWaiting] = None,
        on_skipped: Optional[OnSkipped] = None,
        on_started: Optional[OnStarted] = None,
        on_ready: Optional[OnReady] = None,
        on_retry: Optional[OnRetry] = None,
        on_failed: Optional[OnFailed] = None,
        checkers: Iterable[AnyTask] = [],
        checking_interval: Union[float, int] = 0.05,
        retry: int = 2,
        retry_interval: Union[float, int] = 1,
        max_output_line: int = 1000,
        max_error_line: int = 1000,
        preexec_fn: Optional[Callable[[], Any]] = os.setsid,
        should_execute: Union[bool, str, Callable[..., bool]] = True,
        return_upstream_result: bool = False,
        should_print_cmd_result: bool = True,
        should_show_cmd: bool = True,
        should_show_working_directory: bool = True,
    ):
        BaseTask.__init__(
            self,
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            description=description,
            upstreams=upstreams,
            fallbacks=fallbacks,
            on_triggered=on_triggered,
            on_waiting=on_waiting,
            on_skipped=on_skipped,
            on_started=on_started,
            on_ready=on_ready,
            on_retry=on_retry,
            on_failed=on_failed,
            checkers=checkers,
            checking_interval=checking_interval,
            retry=retry,
            retry_interval=retry_interval,
            should_execute=should_execute,
            return_upstream_result=return_upstream_result,
        )
        max_output_line = max_output_line if max_output_line > 0 else 1
        max_error_line = max_error_line if max_error_line > 0 else 1
        self._cmd = cmd
        self._cmd_path = cmd_path
        self.__set_cwd(cwd)
        self._should_render_cwd = should_render_cwd
        self._max_output_size = max_output_line
        self._max_error_size = max_error_line
        self._output_buffer: Iterable[str] = []
        self._error_buffer: Iterable[str] = []
        if executable is None and DEFAULT_SHELL != "":
            executable = DEFAULT_SHELL
        self._executable = executable
        self._process: Optional[asyncio.subprocess.Process]
        self._preexec_fn = preexec_fn
        self._should_print_cmd_result = should_print_cmd_result
        self._should_show_working_directory = should_show_working_directory
        self._should_show_cmd = should_show_cmd

    def copy(self) -> TCmdTask:
        return super().copy()

    def set_cwd(self, cwd: Union[str, pathlib.Path]):
        self.__set_cwd(cwd)

    def __set_cwd(self, cwd: Optional[Union[str, pathlib.Path]]):
        if cwd is None:
            self._cwd: Union[str, pathlib.Path] = os.getcwd()
            return
        self._cwd: Union[str, pathlib.Path] = os.path.abspath(cwd)

    def to_function(
        self,
        env_prefix: str = "",
        raise_error: bool = True,
        is_async: bool = False,
        show_done_info: bool = True,
        should_clear_xcom: bool = False,
        should_stop_looper: bool = False,
    ) -> Callable[..., CmdResult]:
        return super().to_function(
            env_prefix=env_prefix,
            raise_error=raise_error,
            is_async=is_async,
            show_done_info=show_done_info,
            should_clear_xcom=should_clear_xcom,
            should_stop_looper=should_stop_looper,
        )

    def print_result(self, result: CmdResult):
        if not self._should_print_cmd_result or result.output == "":
            return
        print(result.output)

    async def run(self, *args: Any, **kwargs: Any) -> CmdResult:
        cmd = self.get_cmd_script(*args, **kwargs)
        if self._should_show_cmd:
            self.print_out_dark("Run script: " + self.__get_multiline_repr(cmd))
        cwd = self._get_cwd()
        if self._should_show_working_directory:
            self.print_out_dark("Working directory: " + cwd)
        self._output_buffer = []
        self._error_buffer = []
        process = await asyncio.create_subprocess_shell(
            cmd,
            cwd=cwd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            env=self.get_env_map(),
            shell=True,
            executable=self._executable,
            close_fds=True,
            preexec_fn=self._preexec_fn,
            bufsize=0,
        )
        self._set_task_pid(process.pid)
        self._pids.append(process.pid)
        self._process = process
        try:
            signal.signal(signal.SIGINT, self.__on_kill)
            signal.signal(signal.SIGTERM, self.__on_kill)
        except Exception as e:
            self.print_err(e)
        atexit.register(self.__on_exit)
        await self.__wait_process(process)
        self.log_info("Process completed")
        atexit.unregister(self.__on_exit)
        output = "\n".join(self._output_buffer)
        error = "\n".join(self._error_buffer)
        # get return code
        return_code = process.returncode
        self.log_info(f"Exit status: {return_code}")
        if return_code != 0 and not self._global_state.is_killed_by_signal:
            raise Exception(f"Process {self._name} exited ({return_code}): {error}")
        self.set_task_xcom(key="output", value=output)
        self.set_task_xcom(key="error", value=error)
        return CmdResult(output, error)

    def _get_cwd(self) -> Union[str, pathlib.Path]:
        if self._should_render_cwd and isinstance(self._cwd, str):
            return self.render_str(self._cwd)
        return self._cwd

    def _should_attempt(self) -> bool:
        if self._global_state.no_more_attempt:
            return False
        return super()._should_attempt()

    def _is_last_attempt(self) -> bool:
        if self._global_state.no_more_attempt:
            return True
        return super()._is_last_attempt()

    def __on_kill(self, signum: Any, frame: Any):
        self._global_state.no_more_attempt = True
        self._global_state.is_killed_by_signal = True
        _print_out_dark(f"Getting signal {signum}")
        for pid in self._pids:
            self.__kill_by_pid(pid)
        stop_asyncio_sync()
        _print_out_dark(f"Exiting with signal {signum}")
        sys.exit(signum)

    def __on_exit(self):
        self._global_state.no_more_attempt = True
        self.__kill_by_pid(self._process.pid)

    def __kill_by_pid(self, pid: int):
        """
        Kill a pid, gracefully
        """
        try:
            process_ever_exists = False
            if self.__is_process_exist(pid):
                process_ever_exists = True
                _print_out_dark(f"Send SIGTERM to process {pid}")
                os.killpg(os.getpgid(pid), signal.SIGTERM)
                time.sleep(0.3)
            if self.__is_process_exist(pid):
                _print_out_dark(f"Send SIGINT to process {pid}")
                os.killpg(os.getpgid(pid), signal.SIGINT)
                time.sleep(0.3)
            if self.__is_process_exist(pid):
                _print_out_dark(f"Send SIGKILL to process {pid}")
                os.killpg(os.getpgid(pid), signal.SIGKILL)
            if process_ever_exists:
                _print_out_dark(f"Process {pid} is killed successfully")
        except Exception:
            _log_error(f"Cannot kill process {pid}")

    def __is_process_exist(self, pid: int) -> bool:
        try:
            os.killpg(os.getpgid(pid), 0)
            return True
        except ProcessLookupError:
            return False

    async def __wait_process(self, process: asyncio.subprocess.Process):
        # Create queue
        stdout_queue = asyncio.Queue()
        stderr_queue = asyncio.Queue()
        # Read from streams and put into queue
        stdout_process = asyncio.create_task(
            self.__queue_stream(process.stdout, stdout_queue)
        )
        stderr_process = asyncio.create_task(
            self.__queue_stream(process.stderr, stderr_queue)
        )
        # Handle messages in queue
        stdout_log_process = asyncio.create_task(
            self.__log_from_queue(
                stdout_queue, self.print_out, self._output_buffer, self._max_output_size
            )
        )
        stderr_log_process = asyncio.create_task(
            self.__log_from_queue(
                stderr_queue, self.print_err, self._error_buffer, self._max_error_size
            )
        )
        # wait process
        await process.wait()
        # wait reader and logger
        await stdout_process
        await stderr_process
        await stdout_queue.put(None)
        await stderr_queue.put(None)
        await stdout_log_process
        await stderr_log_process

    def get_cmd_script(self, *args: Any, **kwargs: Any) -> str:
        return self._create_cmd_script(self._cmd_path, self._cmd, *args, **kwargs)

    def _create_cmd_script(
        self, cmd_path: CmdVal, cmd: CmdVal, *args: Any, **kwargs: Any
    ) -> str:
        if not isinstance(cmd_path, str) or cmd_path != "":
            if callable(cmd_path):
                return self.__get_rendered_cmd_path(cmd_path(*args, **kwargs))
            return self.__get_rendered_cmd_path(cmd_path)
        if callable(cmd):
            return self.__get_rendered_cmd(cmd(*args, **kwargs))
        return self.__get_rendered_cmd(cmd)

    def __get_rendered_cmd_path(self, cmd_path: Union[str, Iterable[str]]) -> str:
        if isinstance(cmd_path, str):
            return self.render_file(cmd_path)
        return "\n".join([self.render_file(cmd_path_str) for cmd_path_str in cmd_path])

    def __get_rendered_cmd(
        self, cmd: Union[JinjaTemplate, Iterable[JinjaTemplate]]
    ) -> str:
        if isinstance(cmd, str):
            return self.render_str(cmd)
        return self.render_str("\n".join(list(cmd)))

    async def __queue_stream(self, stream, queue: asyncio.Queue):
        while True:
            try:
                line = await stream.readline()
            except Exception as e:
                line = str(e)
            if not line:
                break
            await queue.put(line)

    async def __log_from_queue(
        self,
        queue: asyncio.Queue,
        print_log: Callable[[str], None],
        buffer: Iterable[str],
        max_size: int,
    ):
        while True:
            line = await queue.get()
            if not line:
                break
            line_str = line.decode("utf-8").rstrip()
            self.__add_to_buffer(buffer, max_size, line_str)
            _reset_stty()
            print_log(line_str)
            _reset_stty()

    def __add_to_buffer(self, buffer: Iterable[str], max_size: int, new_line: str):
        if len(buffer) >= max_size:
            buffer.pop(0)
        buffer.append(new_line)

    def __get_multiline_repr(self, text: str) -> str:
        lines_repr: Iterable[str] = []
        lines = text.split("\n")
        if len(lines) == 1:
            return lines[0]
        for index, line in enumerate(lines):
            line_number_repr = str(index + 1).rjust(4, "0")
            lines_repr.append(f"        {line_number_repr} | {line}")
        return "\n" + "\n".join(lines_repr)



from integration.db_connection import engine
from integration.log import logger
from module.log.entity.activity.repo import ActivityDBRepo, ActivityRepo

activity_repo: ActivityRepo = ActivityDBRepo(logger=logger, engine=engine)



from zrb import DockerComposeStartTask, runner

from ...._project import start_project_containers
from ..._checker import (
    app_container_checker,
    kafka_outside_checker,
    kafka_plaintext_checker,
    pandaproxy_outside_checker,
    pandaproxy_plaintext_checker,
    rabbitmq_checker,
    rabbitmq_management_checker,
    redpanda_console_checker,
)
from ..._constant import PREFER_MICROSERVICES, RESOURCE_DIR
from ..._input import host_input, https_input, local_input
from ...image import build_snake_zrb_app_name_image
from ...image._env import image_env
from ...image._input import image_input
from .._env import compose_env_file, host_port_env
from .._input import enable_monitoring_input
from .._service_config import snake_zrb_app_name_service_configs
from ..remove import remove_snake_zrb_app_name_container
from ._group import snake_zrb_app_name_microservices_container_group
from ._helper import activate_microservices_compose_profile

start_snake_zrb_app_name_microservices_container = DockerComposeStartTask(
    icon="🐳",
    name="start",
    description="Start human readable zrb app name container",
    group=snake_zrb_app_name_microservices_container_group,
    inputs=[
        local_input,
        enable_monitoring_input,
        host_input,
        https_input,
        image_input,
    ],
    should_execute="{{ input.local_snake_zrb_app_name}}",
    upstreams=[build_snake_zrb_app_name_image, remove_snake_zrb_app_name_container],
    cwd=RESOURCE_DIR,
    setup_cmd=activate_microservices_compose_profile,
    compose_env_prefix="CONTAINER_ZRB_ENV_PREFIX",
    compose_service_configs=snake_zrb_app_name_service_configs,
    env_files=[compose_env_file],
    envs=[
        image_env,
        host_port_env,
    ],
    checkers=[
        app_container_checker,
        rabbitmq_checker,
        rabbitmq_management_checker,
        kafka_outside_checker,
        kafka_plaintext_checker,
        redpanda_console_checker,
        pandaproxy_outside_checker,
        pandaproxy_plaintext_checker,
    ],
)

if PREFER_MICROSERVICES:
    start_snake_zrb_app_name_microservices_container >> start_project_containers

runner.register(start_snake_zrb_app_name_microservices_container)



from ._group import snake_zrb_app_name_monolith_container_group
from .start import start_snake_zrb_app_name_monolith_container

assert snake_zrb_app_name_monolith_container_group
assert start_snake_zrb_app_name_monolith_container



from logging import Logger
from typing import Any

from component.messagebus import Publisher
from component.rpc import Caller, Server
from module.log.entity.activity.rpc import register_rpc as register_activity_rpc


def register_rpc(
    logger: Logger, rpc_server: Server, rpc_caller: Caller, publisher: Publisher
):
    logger.info('🥪 Register RPC handlers for "log"')
    register_activity_rpc(logger, rpc_server, rpc_caller, publisher)



from collections.abc import Mapping
from typing import Optional

from component.messagebus.messagebus import Publisher
from module.auth.component import AccessTokenUtil, RefreshTokenUtil
from module.auth.entity.permission.model import PermissionModel
from module.auth.entity.user.repo import UserRepo
from module.auth.schema.token import AccessTokenData, RefreshTokenData, TokenResponse
from module.auth.schema.user import User, UserData, UserLogin, UserResult
from module.log.component.historical_repo_model import HistoricalRepoModel


class UserModel(HistoricalRepoModel[User, UserData, UserResult]):
    schema_result_cls = UserResult
    log_entity_name = "user"

    def __init__(
        self,
        repo: UserRepo,
        publisher: Publisher,
        permission_model: PermissionModel,
        access_token_util: AccessTokenUtil,
        access_token_expire_seconds: int | float,
        refresh_token_util: RefreshTokenUtil,
        refresh_token_expire_seconds: int | float,
        guest_user: User,
        admin_user: Optional[User] = None,
        admin_user_password: str = "",
    ):
        super().__init__(repo, publisher)
        self.permission_model = permission_model
        self.access_token_util = access_token_util
        self.access_token_expire_seconds = access_token_expire_seconds
        self.refresh_token_util = refresh_token_util
        self.refresh_token_expire_seconds = refresh_token_expire_seconds
        self.guest_user = guest_user
        self.admin_user = admin_user
        self.admin_user_pasword = admin_user_password

    async def insert(self, data: UserData) -> User:
        if self.admin_user is not None and data.username == self.admin_user.username:
            raise ValueError(f"Invalid username: {data.username} is used by admin")
        if data.username == self.guest_user.username:
            raise ValueError(f"Invalid username: {data.username} is used by guest")
        return await super().insert(data)

    async def update(self, id: str, data: UserData) -> User:
        if self.admin_user is not None:
            if self.is_admin(id):
                raise ValueError("Forbidden: editting admin user is not permitted")
            if data.username == self.admin_user.username:
                raise ValueError(f"Invalid username: {data.username} is used by admin")
        if self.is_guest(id):
            raise ValueError("Forbidden: editting guest user is not permitted")
        if data.username == self.guest_user.username:
            raise ValueError(f"Invalid username: {data.username} is used by guest")
        return await super().update(id, data)

    async def delete(self, id: str) -> User:
        if self.admin_user is not None and self.is_admin(id):
            raise ValueError("Forbidden: deleting admin user is not permitted")
        if self.is_guest(id):
            raise ValueError("Forbidden: deleting guest user is not permitted")
        return await super().delete(id)

    async def get_by_id(self, id: str) -> User:
        if self.admin_user is not None and self.is_admin(id):
            return self.admin_user
        if id == self.guest_user.id:
            return self.guest_user
        return await super().get_by_id(id)

    async def is_authorized(
        self, id: str, *permission_names: str
    ) -> Mapping[str, bool]:
        user = await self.get_by_id(id)
        if self.is_admin(user.id):
            return {permission_name: True for permission_name in permission_names}
        user_permission_names = self._get_permission_names(user)
        return {
            permission_name: permission_name in user_permission_names
            for permission_name in permission_names
        }

    def _get_permission_names(self, user: User) -> list[str]:
        permission_names = [permission.name for permission in user.permissions]
        for group in user.groups:
            additional_permission_names = [
                permission.name
                for permission in group.permissions
                if permission.name not in permission_names
            ]
            permission_names += additional_permission_names
        return permission_names

    def is_guest(self, id: str) -> bool:
        return id == self.guest_user.id

    def is_admin(self, id: str) -> bool:
        if self.admin_user is None:
            return False
        return id == self.admin_user.id

    async def create_auth_token(self, user_login: UserLogin) -> TokenResponse:
        user = await self._get_user_by_user_login(user_login)
        return TokenResponse(
            access_token=self._get_access_token(user),
            refresh_token=self._get_refresh_token(user),
            token_type="bearer",
        )

    async def refresh_auth_token(
        self, refresh_token: str, access_token: str
    ) -> TokenResponse:
        access_token_data = self.access_token_util.decode(
            access_token, parse_expired_token=True
        )
        refresh_token_data = self.refresh_token_util.decode(refresh_token)
        if access_token_data.user_id != refresh_token_data.user_id:
            raise ValueError("Unmatch refresh and access token")
        user = await self.get_by_id(refresh_token_data.user_id)
        return TokenResponse(
            access_token=self._get_access_token(user),
            refresh_token=self._get_refresh_token(user),
            token_type="bearer",
        )

    def _get_access_token(self, user: User) -> str:
        access_token_data = AccessTokenData(
            user_id=user.id,
            username=user.username,
            expire_seconds=self.access_token_expire_seconds,
        )
        return self.access_token_util.encode(access_token_data)

    def _get_refresh_token(self, user: User) -> str:
        refresh_token_data = RefreshTokenData(
            user_id=user.id, expire_seconds=self.refresh_token_expire_seconds
        )
        return self.refresh_token_util.encode(refresh_token_data)

    async def _get_user_by_user_login(self, user_login: UserLogin) -> User:
        if user_login.identity == "":
            raise ValueError("Invalid identity: Identity is empty")
        if (
            self.admin_user is not None
            and user_login.password == self.admin_user_pasword
            and (
                user_login.identity == self.admin_user.username
                or user_login.identity == self.admin_user.email
                or user_login.identity == self.admin_user.phone
            )
        ):
            return self.admin_user
        return await self.repo.get_by_user_login(user_login)



import asyncio

from zrb.helper.accessories.color import colored
from zrb.helper.log import logger

logger.debug(colored("Loading zrb.helper.asyncio_task", attrs=["dark"]))


async def stop_asyncio():
    tasks = [task for task in asyncio.all_tasks() if task is not asyncio.current_task()]
    for task in tasks:
        task.cancel()

    # Wait until all tasks are cancelled
    await asyncio.gather(*tasks, return_exceptions=True)


def stop_asyncio_sync():
    try:
        loop = asyncio.get_event_loop()
        if loop.is_running():
            loop.create_task(stop_asyncio())
        else:
            loop.run_until_complete(stop_asyncio())
    except Exception:
        pass



from zrb import BoolInput, DockerComposeTask, runner

from ..._project import build_project, build_project_images
from .._constant import RESOURCE_DIR
from .._input import local_input
from ._env import image_env
from ._group import snake_zrb_app_name_image_group
from ._input import image_input

build_snake_zrb_app_name_image = DockerComposeTask(
    icon="🏭",
    name="build",
    description="Build human readable zrb app name image",
    group=snake_zrb_app_name_image_group,
    inputs=[
        local_input,
        image_input,
        BoolInput(
            name="build-kebab-zrb-app-name-with-cache",
            prompt="Build human readable zrb app name image with Cache",
            default=True,
        ),
    ],
    envs=[image_env],
    should_execute="{{ input.local_snake_zrb_app_name}}",
    cwd=RESOURCE_DIR,
    compose_cmd="build",
    compose_args=["kebab-zrb-app-name"],
    compose_flags=[
        "{{ '--no-cache' if not input.build_snake_zrb_app_name_with_cache else '' }}"
    ],
    compose_env_prefix="CONTAINER_ZRB_ENV_PREFIX",
)

build_snake_zrb_app_name_image >> build_project_images
build_snake_zrb_app_name_image >> build_project

runner.register(build_snake_zrb_app_name_image)



import os
from typing import Any

from zrb.builtin.project._helper import (
    create_register_module,
    validate_existing_project_dir,
    validate_inexisting_automation,
)
from zrb.builtin.project._input import project_dir_input
from zrb.builtin.project.add.app._group import project_add_app_group
from zrb.builtin.project.add.app.python._input import (
    app_author_email_input,
    app_author_name_input,
    app_description_input,
    app_image_name_input,
    app_name_input,
    env_prefix_input,
    http_port_input,
)
from zrb.helper.accessories.color import colored
from zrb.helper.util import to_kebab_case
from zrb.runner import runner
from zrb.task.decorator import python_task
from zrb.task.resource_maker import ResourceMaker
from zrb.task.task import Task

_CURRENT_DIR = os.path.dirname(__file__)
_snake_app_name_tpl = "{{util.to_snake_case(input.app_name)}}"


@python_task(
    name="validate",
    inputs=[project_dir_input, app_name_input],
    retry=0,
)
async def validate(*args: Any, **kwargs: Any):
    project_dir = kwargs.get("project_dir")
    validate_existing_project_dir(project_dir)
    app_name = kwargs.get("app_name")
    validate_inexisting_automation(project_dir, app_name)
    source_dir = os.path.join(project_dir, "src", f"{to_kebab_case(app_name)}")
    if os.path.exists(source_dir):
        raise Exception(f"Source already exists: {source_dir}")


copy_resource = ResourceMaker(
    name="copy-resource",
    inputs=[
        project_dir_input,
        app_name_input,
        app_description_input,
        app_author_name_input,
        app_author_email_input,
        app_image_name_input,
        http_port_input,
        env_prefix_input,
    ],
    upstreams=[validate],
    replacements={
        "zrbAppName": "{{input.app_name}}",
        "zrbAppDescription": "{{input.app_description}}",
        "zrbAppAuthorName": "{{input.app_author_name}}",
        "zrbAppAuthorEmail": "{{input.app_author_email}}",
        "zrbAppHttpPort": '{{util.coalesce(input.http_port, "3000")}}',
        "ZRB_ENV_PREFIX": '{{util.coalesce(input.env_prefix, "MY").upper()}}',
        "zrb-app-image-name": "{{input.app_image_name}}",
    },
    template_path=os.path.join(_CURRENT_DIR, "template"),
    destination_path="{{ input.project_dir }}",
    excludes=[
        "*/deployment/venv",
        "*/__pycache__",
    ],
)

register_module = create_register_module(
    module_path=f"_automate.{_snake_app_name_tpl}",
    alias=f"_automate_{_snake_app_name_tpl}",
    inputs=[app_name_input],
    upstreams=[copy_resource],
)


@python_task(
    name="python",
    group=project_add_app_group,
    description="Add simple Python app",
    inputs=[
        project_dir_input,
        app_name_input,
        app_description_input,
        app_author_name_input,
        app_author_email_input,
        app_image_name_input,
        http_port_input,
        env_prefix_input,
    ],
    upstreams=[
        register_module,
    ],
    runner=runner,
)
async def add_python_app(*args: Any, **kwargs: Any):
    task: Task = kwargs.get("_task")
    task.print_out(colored("Python App added", color="yellow"))



import os

from zrb import CmdTask, runner

from .._project import deploy_project
from ._constant import DEPLOYMENT_DIR
from ._env import (
    deployment_app_env_file,
    deployment_config_env_file,
    deployment_replica_env,
    pulumi_backend_url_env,
    pulumi_config_passphrase_env,
)
from ._group import snake_zrb_app_name_group
from ._input import pulumi_stack_input, replica_input
from .image import push_snake_zrb_app_name_image
from .image._env import image_env
from .image._input import image_input

_CURRENT_DIR = os.path.dirname(__file__)

deploy_snake_zrb_app_name = CmdTask(
    icon="🚧",
    name="deploy",
    description="Deploy human readable zrb app name",
    group=snake_zrb_app_name_group,
    inputs=[
        image_input,
        replica_input,
        pulumi_stack_input,
    ],
    upstreams=[push_snake_zrb_app_name_image],
    cwd=DEPLOYMENT_DIR,
    env_files=[
        deployment_config_env_file,
        deployment_app_env_file,
    ],
    envs=[
        pulumi_backend_url_env,
        pulumi_config_passphrase_env,
        image_env,
        deployment_replica_env,
    ],
    cmd_path=[
        os.path.join(_CURRENT_DIR, "init-pulumi-stack.sh"),
        os.path.join(_CURRENT_DIR, "deploy.sh"),
    ],
)

deploy_snake_zrb_app_name >> deploy_project

runner.register(deploy_snake_zrb_app_name)



import asyncio
from collections.abc import Callable, Iterable
from typing import Any, Optional, Union

from zrb.helper.accessories.color import colored
from zrb.helper.accessories.name import get_random_name
from zrb.helper.callable import run_async
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked
from zrb.task.any_task import AnyTask
from zrb.task.any_task_event_handler import (
    OnFailed,
    OnReady,
    OnRetry,
    OnSkipped,
    OnStarted,
    OnTriggered,
    OnWaiting,
)
from zrb.task.checker import Checker
from zrb.task.looper import looper
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_group.group import Group
from zrb.task_input.any_input import AnyInput

logger.debug(colored("Loading zrb.task.watcher", attrs=["dark"]))


@typechecked
class Watcher(Checker):
    __looper = looper

    def __init__(
        self,
        name: str = "watch",
        group: Optional[Group] = None,
        inputs: Iterable[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = "",
        upstreams: Iterable[AnyTask] = [],
        fallbacks: Iterable[AnyTask] = [],
        on_triggered: Optional[OnTriggered] = None,
        on_waiting: Optional[OnWaiting] = None,
        on_skipped: Optional[OnSkipped] = None,
        on_started: Optional[OnStarted] = None,
        on_ready: Optional[OnReady] = None,
        on_retry: Optional[OnRetry] = None,
        on_failed: Optional[OnFailed] = None,
        checking_interval: Union[int, float] = 0.05,
        progress_interval: Union[int, float] = 30,
        expected_result: bool = True,
        should_execute: Union[bool, str, Callable[..., bool]] = True,
    ):
        Checker.__init__(
            self,
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            description=description,
            upstreams=upstreams,
            fallbacks=fallbacks,
            on_triggered=on_triggered,
            on_waiting=on_waiting,
            on_skipped=on_skipped,
            on_started=on_started,
            on_ready=on_ready,
            on_retry=on_retry,
            on_failed=on_failed,
            checking_interval=checking_interval,
            should_execute=should_execute,
            progress_interval=progress_interval,
            expected_result=expected_result,
        )
        self._identifier = get_random_name()

    def get_identifier(self):
        return self._identifier

    async def run(self, *args: Any, **kwargs: Any) -> bool:
        if not looper.is_registered(self._identifier):
            asyncio.create_task(
                run_async(
                    looper.register, self._identifier, self.create_loop_inspector()
                )
            )
        return await super().run(*args, **kwargs)

    async def inspect(self, *args, **kwargs: Any) -> Optional[bool]:
        result = await looper.pop(self._identifier)
        return result

    def create_loop_inspector(self) -> Callable[..., Optional[bool]]:
        def loop_inspect() -> Optional[bool]:
            return None

        return loop_inspect



from component.schema import BaseCountSchema, BaseDateTimeSchema
from module.auth.schema.group import Group
from module.auth.schema.permission import Permission
from pydantic import BaseModel


class UserBase(BaseDateTimeSchema):
    username: str
    phone: str
    email: str
    description: str
    groups: list[str]
    permissions: list[str]


class UserData(UserBase):
    password: str


class User(UserBase):
    id: str
    permissions: list[Permission] = []
    groups: list[Group] = []

    class Config:
        from_attributes = True


class UserResult(BaseCountSchema):
    data: list[User]


class UserLogin(BaseModel):
    identity: str
    password: str



from typing import Any


def activate_all_compose_profile(*args: Any, **kwargs: Any) -> str:
    compose_profile_str = ",".join(
        ["monitoring", "monolith", "microservices", "kafka", "postgres", "rabbitmq"]
    )
    return f"export COMPOSE_PROFILES={compose_profile_str}"



from _common import (
    NAMESPACE,
    POSTGRESQL_AUTH_PASSWORD,
    POSTGRESQL_AUTH_POSTGRES_PASSWORD,
    POSTGRESQL_AUTH_USERNAME,
    POSTGRESQL_DB,
)
from pulumi_kubernetes.helm.v3 import Chart, LocalChartOpts


def create_postgresql() -> Chart:
    postgresql = Chart(
        "postgresql",
        LocalChartOpts(
            path="./helm-charts/postgresql",
            namespace=NAMESPACE,
            values={
                "auth": {
                    "postgresPassword": POSTGRESQL_AUTH_POSTGRES_PASSWORD,
                    "username": POSTGRESQL_AUTH_USERNAME,
                    "password": POSTGRESQL_AUTH_PASSWORD,
                    "database": POSTGRESQL_DB,
                },
                "resources": {
                    "limits": {"cpu": "250m", "memory": "256Mi"},
                    "requests": {"cpu": "250m", "memory": "156Mi"},
                },
            },
            skip_await=True,
        ),
    )
    return postgresql



import os

from zrb import CmdTask, runner

from ..._project import start_project
from .._checker import app_local_checker
from .._constant import APP_DIR, AUTOMATE_DIR, PREFER_MICROSERVICES
from .._env import app_enable_otel_env, app_env_file
from .._input import host_input, https_input, local_input
from ..backend import prepare_snake_zrb_app_name_backend
from ..container._input import enable_monitoring_input
from ..container.support import start_snake_zrb_app_name_support_container
from ..frontend import build_snake_zrb_app_name_frontend
from ._group import snake_zrb_app_name_monolith_group

_CURRENT_DIR = os.path.dirname(__file__)

start_snake_zrb_app_name_monolith = CmdTask(
    icon="🚤",
    name="start",
    inputs=[
        local_input,
        enable_monitoring_input,
        host_input,
        https_input,
    ],
    description="Start human readable zrb app name as a monolith",
    group=snake_zrb_app_name_monolith_group,
    should_execute="{{ input.local_snake_zrb_app_name}}",
    upstreams=[
        start_snake_zrb_app_name_support_container,
        build_snake_zrb_app_name_frontend,
        prepare_snake_zrb_app_name_backend,
    ],
    cwd=APP_DIR,
    env_files=[app_env_file],
    envs=[app_enable_otel_env],
    cmd_path=[
        os.path.join(AUTOMATE_DIR, "activate-venv.sh"),
        os.path.join(_CURRENT_DIR, "start.sh"),
    ],
    checkers=[
        app_local_checker,
    ],
)

if not PREFER_MICROSERVICES:
    start_snake_zrb_app_name_monolith >> start_project

runner.register(start_snake_zrb_app_name_monolith)



from zrb.helper.accessories.name import get_random_name
from zrb.task_input.str_input import StrInput

task_name_input = StrInput(
    name="task-name",
    shortcut="t",
    description="Task name",
    prompt="Task name",
    default=f"run-{get_random_name()}",
)



from zrb.builtin.ubuntu._group import ubuntu_group
from zrb.runner import runner
from zrb.task.cmd_task import CmdTask

update_ubuntu = CmdTask(
    name="update",
    group=ubuntu_group,
    description="Update ubuntu",
    cmd=[
        "sudo apt update",
        "sudo apt upgrade -y",
    ],
    retry_interval=3,
    preexec_fn=None,
)
runner.register(update_ubuntu)



from zrb.builtin.git._group import git_group
from zrb.runner import runner
from zrb.task.cmd_task import CmdTask

clear_git_branch = CmdTask(
    name="clear-branch",
    group=git_group,
    description="Clear branches",
    cmd=[
        "for BRANCH in $(git branch)",
        "do",
        '  if [ "$BRANCH" != "main" ] && [ "$BRANCH" != "*" ]',
        "  then",
        "    git branch -D $BRANCH",
        "  fi",
        "done",
    ],
)
runner.register(clear_git_branch)



from typing import Any, Optional, Union

from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked
from zrb.task_input.base_input import BaseInput, InputCallback, InputDefault

logger.debug(colored("Loading zrb.task_input.int_input", attrs=["dark"]))

# flake8: noqa E501


@typechecked
class IntInput(BaseInput):
    """
    A specialized input class for handling integer values in task inputs.

    `IntInput` extends `BaseInput` to specifically handle inputs where integer values are required.
    It supports various features like default values, prompts, flags, and other customization options,
    making it suitable for tasks that require numerical input in the form of integers.

    Args:
        name (str): The name of the input, serving as an identifier.
        shortcut (Optional[str]): An optional shortcut for easier reference to the input.
        default (Optional[Any]): The default value for the input, should be an integer if provided.
        callback (Optional[Any]): The default value of the input.
        description (Optional[str]): A brief description of what the input is for.
        show_default (Union[bool, JinjaTemplate, None]): Determines the default value to be shown.
        prompt (Union[bool, str]): A boolean or string to determine the prompt for user input. If `True`, uses a default prompt.
        confirmation_prompt (Union[bool, str]): If `True`, the user will be asked to confirm their input.
        prompt_required (bool): If `True`, makes the input prompt mandatory.
        hide_input (bool): If `True`, hides the input value, typically used for sensitive information.
        is_flag (Optional[bool]): Indicates if the input functions as a flag, taking boolean values.
        flag_value (Optional[Any]): The value associated with the input when used as a flag.
        multiple (bool): If `True`, allows entering multiple integer values.
        count (bool): If `True`, counts the occurrences of the input.
        allow_from_autoenv (bool): If `True`, enables automatic population of the input from environment variables.
        hidden (bool): If `True`, keeps the input hidden in help messages or documentation.
        show_choices (bool): If `True`, shows any restricted choices for the input value.
        show_envvar (bool): If `True`, displays the associated environment variable, if applicable.
        nargs (int): The number of arguments that the input can accept.
        should_render (bool): If `True`, renders the input in the user interface or command-line interface.

    Examples:
        >>> int_input = IntInput(name='age', default=30, description='Enter your age')
        >>> int_input.get_default()
        30
    """

    def __init__(
        self,
        name: str,
        shortcut: Optional[str] = None,
        default: Optional[Union[Any, InputDefault]] = None,
        callback: Optional[InputCallback] = None,
        description: Optional[str] = None,
        show_default: Union[bool, str] = True,
        prompt: Union[bool, str] = True,
        confirmation_prompt: Union[bool, str] = False,
        prompt_required: bool = True,
        hide_input: bool = False,
        is_flag: Optional[bool] = None,
        flag_value: Optional[Any] = None,
        multiple: bool = False,
        count: bool = False,
        allow_from_autoenv: bool = True,
        hidden: bool = False,
        show_choices: bool = True,
        show_envvar: bool = False,
        nargs: int = 1,
        should_render: bool = True,
    ):
        BaseInput.__init__(
            self,
            name=name,
            shortcut=shortcut,
            default=default,
            callback=callback,
            description=description,
            show_default=show_default,
            prompt=prompt,
            confirmation_prompt=confirmation_prompt,
            prompt_required=prompt_required,
            hide_input=hide_input,
            is_flag=is_flag,
            flag_value=flag_value,
            multiple=multiple,
            count=count,
            allow_from_autoenv=allow_from_autoenv,
            type=int,
            hidden=hidden,
            show_choices=show_choices,
            show_envvar=show_envvar,
            nargs=nargs,
            should_render=should_render,
        )



import os
from typing import Any, Optional

from zrb.builtin.project._input import project_dir_input
from zrb.helper import util
from zrb.helper.accessories.color import colored
from zrb.helper.codemod.add_assert_resource import add_assert_resource
from zrb.helper.codemod.add_import_module import add_import_module
from zrb.helper.file.text import read_text_file_async, write_text_file_async
from zrb.helper.typecheck import typechecked
from zrb.task.any_task import AnyTask
from zrb.task.decorator import python_task
from zrb.task.task import Task
from zrb.task_input.any_input import AnyInput


@typechecked
def validate_existing_project_dir(project_dir: str):
    if not os.path.isfile(os.path.join(project_dir, "zrb_init.py")):
        raise Exception(f"Not a project: {project_dir}")


@typechecked
def validate_inexisting_automation(project_dir: str, automation_name: str):
    validate_inexisting_automation_package(project_dir, automation_name)
    validate_inexisting_automation_module(project_dir, automation_name)


@typechecked
def validate_inexisting_automation_package(project_dir: str, package_name: str):
    package_dir = os.path.join(
        project_dir, "_automate", f"{util.to_snake_case(package_name)}"
    )
    if os.path.exists(package_dir):
        raise Exception(f"Automation package already exists: {package_dir}")


@typechecked
def validate_inexisting_automation_module(project_dir: str, module_name: str):
    module_file = os.path.join(
        project_dir, "_automate", f"{util.to_snake_case(module_name)}.py"
    )
    if os.path.exists(module_file):
        raise Exception(f"Automation module already exists: {module_file}")


@typechecked
async def register_module_to_project(
    project_dir: str, module_path: str, alias: Optional[str] = None
):
    zrb_init_path = os.path.join(project_dir, "zrb_init.py")
    code = await read_text_file_async(zrb_init_path)
    if alias is None:
        alias = module_path.split(".")[-1]
    code = add_import_module(code=code, module_path=module_path, alias=alias)
    code = add_assert_resource(code, alias)
    await write_text_file_async(zrb_init_path, code)
    return True


@typechecked
def create_register_module(
    module_path: str,
    inputs: Optional[list[AnyInput]] = None,
    upstreams: Optional[list[AnyTask]] = None,
    alias: Optional[str] = None,
    task_name: str = "register-module",
) -> Task:
    @python_task(
        name=task_name,
        inputs=[project_dir_input] + inputs if inputs is not None else [],
        upstreams=upstreams if upstreams is not None else [],
    )
    async def register_module(*args: Any, **kwargs: Any):
        task: Task = kwargs.get("_task")
        project_dir = kwargs.get("project_dir")
        validate_existing_project_dir(project_dir)
        rendered_module_path = task.render_str(module_path)
        rendered_alias: Optional[str] = None
        if alias is not None:
            rendered_alias = task.render_str(alias)
        module_import_label = _get_module_import_label(
            module_path=rendered_module_path, alias=rendered_alias
        )
        task.print_out(colored(module_import_label, color="yellow"))
        await register_module_to_project(
            project_dir=project_dir,
            module_path=rendered_module_path,
            alias=rendered_alias,
        )

    return register_module


def _get_module_import_label(module_path: str, alias: Optional[str]):
    if alias == "" or alias is None:
        return module_path
    return f"{module_path} as {alias}"



from component.messagebus import (
    Admin,
    Consumer,
    KafkaAdmin,
    KafkaConsumer,
    KafkaPublisher,
    MessageSerializer,
    MockAdmin,
    MockConsumer,
    MockPublisher,
    Publisher,
    RMQAdmin,
    RMQConsumer,
    RMQPublisher,
)
from config import (
    APP_BROKER_TYPE,
    APP_KAFKA_BOOTSTRAP_SERVERS,
    APP_KAFKA_SASL_MECHANISM,
    APP_KAFKA_SASL_PASS,
    APP_KAFKA_SASL_USER,
    APP_KAFKA_SECURITY_PROTOCOL,
    APP_NAME,
    APP_RMQ_CONNECTION_STRING,
)
from integration.log import logger


def init_admin(default_admin: Admin) -> Admin:
    if APP_BROKER_TYPE == "rabbitmq":
        return RMQAdmin(
            logger=logger, connection_string=APP_RMQ_CONNECTION_STRING, configs={}
        )
    if APP_BROKER_TYPE == "kafka":
        return KafkaAdmin(
            logger=logger,
            bootstrap_servers=APP_KAFKA_BOOTSTRAP_SERVERS,
            security_protocol=APP_KAFKA_SECURITY_PROTOCOL,
            sasl_mechanism=APP_KAFKA_SASL_MECHANISM,
            sasl_plain_username=APP_KAFKA_SASL_USER,
            sasl_plain_password=APP_KAFKA_SASL_PASS,
            configs={},
        )
    return default_admin


def init_publisher(
    serializer: MessageSerializer, admin: Admin, default_publisher: Publisher
) -> Publisher:
    if APP_BROKER_TYPE == "rabbitmq":
        return RMQPublisher(
            logger=logger,
            connection_string=APP_RMQ_CONNECTION_STRING,
            serializer=serializer,
            rmq_admin=admin,
        )
    if APP_BROKER_TYPE == "kafka":
        return KafkaPublisher(
            logger=logger,
            bootstrap_servers=APP_KAFKA_BOOTSTRAP_SERVERS,
            security_protocol=APP_KAFKA_SECURITY_PROTOCOL,
            sasl_mechanism=APP_KAFKA_SASL_MECHANISM,
            sasl_plain_username=APP_KAFKA_SASL_USER,
            sasl_plain_password=APP_KAFKA_SASL_PASS,
            serializer=serializer,
            kafka_admin=admin,
        )
    return default_publisher


def init_consumer(
    serializer: MessageSerializer, admin: Admin, default_consumer: Consumer
) -> Consumer:
    if APP_BROKER_TYPE == "rabbitmq":
        return RMQConsumer(
            logger=logger,
            connection_string=APP_RMQ_CONNECTION_STRING,
            serializer=serializer,
            rmq_admin=admin,
        )
    if APP_BROKER_TYPE == "kafka":
        return KafkaConsumer(
            logger=logger,
            bootstrap_servers=APP_KAFKA_BOOTSTRAP_SERVERS,
            security_protocol=APP_KAFKA_SECURITY_PROTOCOL,
            sasl_mechanism=APP_KAFKA_SASL_MECHANISM,
            sasl_plain_username=APP_KAFKA_SASL_USER,
            sasl_plain_password=APP_KAFKA_SASL_PASS,
            group_id=APP_NAME,
            serializer=serializer,
            kafka_admin=admin,
        )
    return default_consumer


def init_message_serializer() -> MessageSerializer:
    # Add custom logic if necessary
    return MessageSerializer()


message_serializer = init_message_serializer()
mock_admin = MockAdmin()
mock_consumer = MockConsumer(logger=logger, serializer=message_serializer)
mock_publisher = MockPublisher(
    logger=logger, consumer=mock_consumer, serializer=message_serializer
)
admin = init_admin(default_admin=mock_admin)
consumer = init_consumer(
    serializer=message_serializer, admin=admin, default_consumer=mock_consumer
)
publisher = init_publisher(
    serializer=message_serializer, admin=admin, default_publisher=mock_publisher
)



from collections.abc import Mapping
from logging import Logger
from typing import Any

from component.messagebus import Publisher
from component.repo import SearchFilter
from component.rpc import Caller, Server
from module.auth.integration.model.permission_model import permission_model
from module.auth.schema.permission import PermissionData
from module.auth.schema.token import AccessTokenData


def register_rpc(
    logger: Logger, rpc_server: Server, rpc_caller: Caller, publisher: Publisher
):
    logger.info('🥪 Register RPC handlers for "auth.permission"')

    @rpc_server.register("auth_ensure_permission")
    async def ensure(data: Mapping[str, Any]):
        await permission_model.ensure_permission(PermissionData(**data))

    @rpc_server.register("auth_get_permission")
    async def get(
        keyword: str,
        criterion: Mapping[str, Any],
        limit: int,
        offset: int,
        user_token_data: Mapping[str, Any],
    ) -> Mapping[str, Any]:
        result = await permission_model.get(
            search_filter=SearchFilter(keyword=keyword, criterion=criterion),
            limit=limit,
            offset=offset,
        )
        return result.model_dump()

    @rpc_server.register("auth_get_permission_by_id")
    async def get_by_id(
        id: str, user_token_data: Mapping[str, Any] = {}
    ) -> Mapping[str, Any]:
        row = await permission_model.get_by_id(id)
        return row.model_dump()

    @rpc_server.register("auth_insert_permission")
    async def insert(
        data: Mapping[str, Any], user_token_data: Mapping[str, Any]
    ) -> Mapping[str, Any]:
        user_token_data = AccessTokenData(**user_token_data)
        data["created_by"] = user_token_data.user_id
        data["updated_by"] = user_token_data.user_id
        row = await permission_model.insert(data=PermissionData(**data))
        return row.model_dump()

    @rpc_server.register("auth_update_permission")
    async def update(
        id: str, data: Mapping[str, Any], user_token_data: Mapping[str, Any]
    ) -> Mapping[str, Any]:
        user_token_data = AccessTokenData(**user_token_data)
        data["updated_by"] = user_token_data.user_id
        row = await permission_model.update(id=id, data=PermissionData(**data))
        return row.model_dump()

    @rpc_server.register("auth_delete_permission")
    async def delete(id: str, user_token_data: Mapping[str, Any]) -> Mapping[str, Any]:
        user_token_data = AccessTokenData(**user_token_data)
        row = await permission_model.delete(id=id)
        return row.model_dump()



import re
from collections.abc import Mapping

from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked

logger.debug(colored("Loading zrb.helper.string.parse_replacment", attrs=["dark"]))


@typechecked
def parse_replacement(text: str, replacement: Mapping[str, str]):
    new_text = text
    for old, new in replacement.items():
        if len(new.strip().split("\n")) > 1:
            pattern = "".join(["^([ \\t]+)", re.escape(old)])
            new_replacement = "".join(["\\1", new.replace("\n", "\n\\1")])
            new_text = re.sub(pattern, new_replacement, new_text, flags=re.MULTILINE)
            continue
        new_text = new_text.replace(old, new)
    return new_text



import os

from zrb import CmdTask, runner

from .._constant import APP_DIR, AUTOMATE_DIR
from ._group import snake_zrb_app_name_backend_group

_CURRENT_DIR = os.path.dirname(__file__)

prepare_snake_zrb_app_name_backend = CmdTask(
    icon="🚤",
    name="prepare",
    description="Prepare backend for human readable zrb app name",
    group=snake_zrb_app_name_backend_group,
    cwd=APP_DIR,
    cmd_path=[
        os.path.join(AUTOMATE_DIR, "activate-venv.sh"),
        os.path.join(_CURRENT_DIR, "prepare.sh"),
    ],
)
runner.register(prepare_snake_zrb_app_name_backend)



from collections.abc import Mapping
from typing import Any, Optional

from fastapi import HTTPException


class HTTPAPIException(HTTPException):
    def __init__(
        self,
        status_code: Optional[int] = None,
        error_message: Optional[str] = None,
        headers: Optional[Mapping[str, Any]] = None,
        error: Optional[Exception] = None,
    ):
        if error is HTTPException:
            status_code = (error.status_code,)
            detail = (error.detail,)
            headers = self._get_headers(error.headers)
            return super().__init__(
                status_code=status_code, detail=detail, headers=headers
            )
        if error is not None:
            error_message = f"{error}"
        status_code = self._get_status_code(status_code, error_message)
        headers = self._get_headers(headers)
        return super().__init__(
            status_code=status_code, detail=error_message, headers=headers
        )

    def _get_status_code(
        self, status_code: Optional[int], error_message: Optional[str]
    ) -> int:
        if status_code is not None:
            return status_code
        if error_message is None:
            return 500
        if error_message.lower().startswith("not found"):
            return 404
        if error_message.lower().startswith("forbidden"):
            return 403
        if error_message.lower().startswith("unauthorized"):
            return 403
        if error_message.lower().startswith("unauthenticated"):
            return 401
        if error_message.lower().startswith("unprocessable"):
            return 422
        if error_message.lower().startswith("invalid"):
            return 422
        return 500

    def _get_headers(
        self, original_headers: Optional[Mapping[str, Any]] = None
    ) -> Mapping[str, Any]:
        headers = {} if original_headers is None else dict(original_headers)
        headers["api-error"] = "yes"
        return headers



import os

from zrb import CmdTask, HTTPChecker, Task
from zrb.helper.util import to_kebab_case

from .._constant import APP_DIR, AUTOMATE_DIR, MODULES
from .._input import host_input, https_input, local_input
from ..backend import prepare_snake_zrb_app_name_backend
from ..container._input import enable_monitoring_input
from ..container.support import start_snake_zrb_app_name_support_container
from ..frontend import build_snake_zrb_app_name_frontend
from ._helper import get_disable_all_module_envs, get_service_env_file, get_service_envs

_CURRENT_DIR = os.path.dirname(__file__)


_disable_all_module_envs = get_disable_all_module_envs()
start_snake_zrb_app_name_services: list[Task] = []
for _module_index, _module_name in enumerate(MODULES):
    _service_env_file = get_service_env_file(_module_name)
    _service_envs = get_service_envs(zrbAppHttpPort, _module_index, _module_name)
    _kebab_module_name = to_kebab_case(_module_name)
    # Define start service task
    _start_service = CmdTask(
        name=f"start-{_kebab_module_name}-service",
        inputs=[
            local_input,
            host_input,
            https_input,
            enable_monitoring_input,
        ],
        upstreams=[
            start_snake_zrb_app_name_support_container,
            build_snake_zrb_app_name_frontend,
            prepare_snake_zrb_app_name_backend,
        ],
        cwd=APP_DIR,
        env_files=[_service_env_file],
        envs=[*_disable_all_module_envs, *_service_envs],
        cmd_path=[
            os.path.join(AUTOMATE_DIR, "activate-venv.sh"),
            os.path.join(_CURRENT_DIR, "start.sh"),
        ],
        checkers=[
            HTTPChecker(
                name=f"check-kebab-zrb-app-name-{_kebab_module_name}-service",
                host="{{input.snake_zrb_app_name_host}}",
                url="/readiness",
                port="{{env.APP_PORT}}",
                is_https="{{input.snake_zrb_app_name_https}}",
            )
        ],
    )
    start_snake_zrb_app_name_services.append(_start_service)



from collections.abc import Callable, Iterable
from typing import Any, Optional, TypeVar, Union

from zrb.helper.accessories.color import colored
from zrb.helper.file.match import get_file_names
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked
from zrb.helper.typing import JinjaTemplate
from zrb.task.any_task import AnyTask
from zrb.task.any_task_event_handler import (
    OnFailed,
    OnReady,
    OnRetry,
    OnSkipped,
    OnStarted,
    OnTriggered,
    OnWaiting,
)
from zrb.task.checker import Checker
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_group.group import Group
from zrb.task_input.any_input import AnyInput

logger.debug(colored("Loading zrb.task.path_checker", attrs=["dark"]))

TPathChecker = TypeVar("TPathChecker", bound="PathChecker")


@typechecked
class PathChecker(Checker):
    def __init__(
        self,
        name: str = "check-path",
        group: Optional[Group] = None,
        inputs: Iterable[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = "",
        upstreams: Iterable[AnyTask] = [],
        fallbacks: Iterable[AnyTask] = [],
        on_triggered: Optional[OnTriggered] = None,
        on_waiting: Optional[OnWaiting] = None,
        on_skipped: Optional[OnSkipped] = None,
        on_started: Optional[OnStarted] = None,
        on_ready: Optional[OnReady] = None,
        on_retry: Optional[OnRetry] = None,
        on_failed: Optional[OnFailed] = None,
        path: JinjaTemplate = "",
        ignored_path: Union[JinjaTemplate, Iterable[JinjaTemplate]] = [],
        checking_interval: Union[int, float] = 0.05,
        progress_interval: Union[int, float] = 5,
        expected_result: bool = True,
        should_execute: Union[bool, JinjaTemplate, Callable[..., bool]] = True,
    ):
        Checker.__init__(
            self,
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            description=description,
            upstreams=upstreams,
            fallbacks=fallbacks,
            on_triggered=on_triggered,
            on_waiting=on_waiting,
            on_skipped=on_skipped,
            on_started=on_started,
            on_ready=on_ready,
            on_retry=on_retry,
            on_failed=on_failed,
            checking_interval=checking_interval,
            progress_interval=progress_interval,
            expected_result=expected_result,
            should_execute=should_execute,
        )
        self._path = path
        self._ignored_path = ignored_path
        self._rendered_path: str = ""
        self._rendered_ignored_paths: list[str] = []

    def copy(self) -> TPathChecker:
        return super().copy()

    def to_function(
        self,
        env_prefix: str = "",
        raise_error: bool = True,
        is_async: bool = False,
        show_done_info: bool = True,
        should_clear_xcom: bool = False,
        should_stop_looper: bool = False,
    ) -> Callable[..., bool]:
        return super().to_function(
            env_prefix=env_prefix,
            raise_error=raise_error,
            is_async=is_async,
            show_done_info=show_done_info,
            should_clear_xcom=should_clear_xcom,
            should_stop_looper=should_stop_looper,
        )

    async def run(self, *args: Any, **kwargs: Any) -> bool:
        self._rendered_path = self.render_str(self._path)
        self._rendered_ignored_paths = [
            ignored_path
            for ignored_path in self._get_rendered_ignored_paths()
            if ignored_path != ""
        ]
        return await super().run(*args, **kwargs)

    def _get_rendered_ignored_paths(self) -> list[str]:
        if isinstance(self._ignored_path, str):
            return [self.render_str(self._ignored_path)]
        return [self.render_str(ignored_path) for ignored_path in self._ignored_path]

    async def inspect(self, *args: Any, **kwargs: Any) -> bool:
        label = f"Checking {self._rendered_path}"
        try:
            matches = get_file_names(
                glob_path=self._rendered_path,
                glob_ignored_paths=self._rendered_ignored_paths,
            )
            if len(matches) > 0:
                self.print_out(f"{label} (Exist)")
                return True
            self.show_progress(f"{label} (Not Exist)")
        except Exception:
            self.show_progress(f"{label} Cannot inspect")
        return False



import sys

from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.task import show_lines as task_show_lines
from zrb.helper.typecheck import typechecked
from zrb.task.task import Task

logger.debug(colored("Loading zrb.helper.python_task", attrs=["dark"]))

_DEPRECATION_WARNING = """
DEPRECATED: zrb.helper.python_task
Use zrb.helper.task instead

```python
from zrb.helper.task import show_lines
```
"""

print(colored(_DEPRECATION_WARNING, color="red", attrs=["bold"]), file=sys.stderr)


@typechecked
def show_lines(task: Task, *lines: str):
    task_show_lines(task, *lines)



import os

from zrb import Env, EnvFile

from .._constant import RESOURCE_DIR

compose_env_file = EnvFile(
    path=os.path.join(RESOURCE_DIR, "docker-compose.env"),
    prefix="CONTAINER_ZRB_ENV_PREFIX",
)

host_port_env = Env(
    name="HOST_PORT",
    os_name="CONTAINER_ZRB_ENV_PREFIX_HOST_PORT",
    default="zrbAppHttpPort",
)



from zrb import Group
from zrb.builtin.project.add._group import project_add_group

project_add_fastapp_group = Group(
    name="fastapp",
    description="Add fastapp related resources",
    parent=project_add_group,
)



from zrb.builtin.base64 import base64_group, decode_base64, encode_base64
from zrb.builtin.devtool import (
    devtool_group,
    devtool_install_group,
    install_aws,
    install_docker,
    install_gcloud,
    install_gvm,
    install_helix,
    install_helm,
    install_kubectl,
    install_nvm,
    install_pulumi,
    install_pyenv,
    install_sdkman,
    install_selenium,
    install_terraform,
    install_tmux,
    install_zsh,
)
from zrb.builtin.docker import docker_group, prune_docker
from zrb.builtin.env import env_group, get_env
from zrb.builtin.eval import evaluate
from zrb.builtin.explain import explain_group, explain_tasks
from zrb.builtin.git import clear_git_branch, get_git_file_changes, git_group
from zrb.builtin.md5 import hash_text_md5, md5_group, sum_file_md5
from zrb.builtin.monorepo import add_to_monorepo, pull_from_monorepo, push_to_monorepo
from zrb.builtin.process import (
    get_process_pid_by_name,
    get_process_pid_by_port,
    process_group,
    process_pid_group,
)
from zrb.builtin.project import (
    add_cmd_task,
    add_docker_compose_task,
    add_fastapp_application,
    add_fastapp_crud,
    add_fastapp_field,
    add_fastapp_module,
    add_plugin,
    add_project_tasks,
    add_python_app,
    add_python_task,
    create_project,
    project_add_app_group,
    project_add_fastapp_group,
    project_add_group,
    project_add_task_group,
    project_group,
)
from zrb.builtin.say import say
from zrb.builtin.schedule import schedule
from zrb.builtin.ubuntu import (
    install_ubuntu_all,
    install_ubuntu_essentials,
    install_ubuntu_tex,
    install_ubuntu_toys,
    ubuntu_group,
    ubuntu_install_group,
    update_ubuntu,
)
from zrb.builtin.update import update_zrb
from zrb.builtin.version import get_version
from zrb.builtin.watch_changes import watch_changes

assert base64_group
assert decode_base64
assert encode_base64
assert devtool_group
assert devtool_install_group
assert install_aws
assert install_docker
assert install_gcloud
assert install_gvm
assert install_helix
assert install_helm
assert install_kubectl
assert install_nvm
assert install_pulumi
assert install_pyenv
assert install_sdkman
assert install_selenium
assert install_terraform
assert install_tmux
assert install_zsh
assert docker_group
assert prune_docker
assert env_group
assert get_env
assert git_group
assert clear_git_branch
assert get_git_file_changes
assert md5_group
assert hash_text_md5
assert sum_file_md5
assert process_group
assert process_pid_group
assert get_process_pid_by_name
assert get_process_pid_by_port
assert project_group
assert project_add_group
assert project_add_app_group
assert project_add_fastapp_group
assert project_add_task_group
assert create_project
assert add_plugin
assert add_fastapp_application
assert add_fastapp_crud
assert add_fastapp_module
assert add_fastapp_field
assert add_python_app
assert add_project_tasks
assert add_cmd_task
assert add_docker_compose_task
assert add_python_task
assert add_to_monorepo
assert pull_from_monorepo
assert push_to_monorepo
assert ubuntu_group
assert ubuntu_install_group
assert update_ubuntu
assert ubuntu_install_group
assert install_ubuntu_all
assert install_ubuntu_essentials
assert install_ubuntu_tex
assert install_ubuntu_toys
assert explain_group
assert explain_tasks
assert evaluate
assert say
assert schedule
assert update_zrb
assert get_version
assert watch_changes



import os

MESSAGE = os.getenv("APP_MESSAGE", "Hello, world!")
PORT = int(os.getenv("APP_PORT", "8000"))


async def app(scope, receive, send):
    assert scope["type"] == "http"

    await send(
        {
            "type": "http.response.start",
            "status": 200,
            "headers": [
                [b"content-type", b"text/plain"],
            ],
        }
    )
    await send(
        {
            "type": "http.response.body",
            "body": bytes(MESSAGE, "utf-8"),
        }
    )



import os

from zrb import EnvFile, ServiceConfig

from .._constant import APP_DIR

snake_zrb_app_name_service_config = ServiceConfig(
    env_files=[
        EnvFile(
            path=os.path.join(APP_DIR, "template.env"),
            prefix="CONTAINER_ZRB_ENV_PREFIX",
        )
    ]
)



import os
from collections.abc import Callable, Iterable, Mapping
from typing import Any, Optional, Union

from zrb.action.runner import Runner
from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.task import show_lines
from zrb.helper.typecheck import typechecked
from zrb.helper.util import to_human_readable
from zrb.task.any_task import AnyTask
from zrb.task.any_task_event_handler import (
    OnFailed,
    OnReady,
    OnRetry,
    OnSkipped,
    OnStarted,
    OnTriggered,
    OnWaiting,
)
from zrb.task.task import Task
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_group.group import Group
from zrb.task_input.any_input import AnyInput

logger.debug(colored("Loading zrb.task.wiki_task", attrs=["dark"]))


@typechecked
def create_wiki_tasks(
    directory: str,
    group: Optional[Group] = None,
    inputs: Iterable[AnyInput] = [],
    envs: Iterable[Env] = [],
    env_files: Iterable[EnvFile] = [],
    icon: Optional[str] = None,
    color: Optional[str] = None,
    upstreams: Iterable[AnyTask] = [],
    fallbacks: Iterable[AnyTask] = [],
    on_triggered: Optional[OnTriggered] = None,
    on_waiting: Optional[OnWaiting] = None,
    on_skipped: Optional[OnSkipped] = None,
    on_started: Optional[OnStarted] = None,
    on_ready: Optional[OnReady] = None,
    on_retry: Optional[OnRetry] = None,
    on_failed: Optional[OnFailed] = None,
    should_execute: Union[bool, str, Callable[..., bool]] = True,
    runner: Optional[Runner] = None,
):
    abs_directory = os.path.abspath(directory)
    directory_structure = _get_directory_structure(abs_directory)
    tasks: list[AnyTask] = []
    for file_name in directory_structure["files"]:
        if not file_name.endswith(".md"):
            continue
        task_name = file_name[:-3]
        task_description = f"Article about {to_human_readable(task_name)}"
        task = Task(
            name=task_name,
            group=group,
            description=task_description,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            upstreams=upstreams,
            fallbacks=fallbacks,
            on_triggered=on_triggered,
            on_waiting=on_waiting,
            on_skipped=on_skipped,
            on_started=on_started,
            on_ready=on_ready,
            on_retry=on_retry,
            on_failed=on_failed,
            should_execute=should_execute,
            run=_create_function(directory=directory, file_name=file_name),
        )
        if runner is not None:
            runner.register(task)
        tasks.append(task)
    for dir_name in directory_structure["dirs"]:
        sub_group_description = f"Articles related to {to_human_readable(dir_name)}"
        sub_group = Group(
            name=dir_name,
            parent=group,
            description=sub_group_description,
        )
        sub_tasks = create_wiki_tasks(
            directory=os.path.join(abs_directory, dir_name),
            group=sub_group,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            upstreams=upstreams,
            fallbacks=fallbacks,
            on_triggered=on_triggered,
            on_waiting=on_waiting,
            on_skipped=on_skipped,
            on_started=on_started,
            on_ready=on_ready,
            on_retry=on_retry,
            on_failed=on_failed,
            should_execute=should_execute,
            runner=runner,
        )
        tasks = tasks + sub_tasks
    return tasks


def _create_function(directory: str, file_name: str) -> Callable[..., Any]:
    def fn(*args: Any, **kwargs: Any):
        with open(os.path.join(directory, file_name)) as f:
            content = f.read()
            lines = content.split("\n")
            show_lines(kwargs["_task"], *lines)

    return fn


def _get_directory_structure(path) -> Mapping[str, list[str]]:
    contents = {"files": [], "dirs": []}
    for item in os.listdir(path):
        full_path = os.path.join(path, item)
        if os.path.isdir(full_path):
            contents["dirs"].append(item)
        elif os.path.isfile(full_path):
            contents["files"].append(item)
    return contents



import os

from zrb.helper.util import to_kebab_case


def get_app_module_dir(project_dir: str, app_name: str) -> str:
    kebab_app_name = to_kebab_case(app_name)
    return os.path.join(project_dir, "src", kebab_app_name, "src", "module")



from asgi_lifespan import LifespanManager
from httpx import AsyncClient
from src.main import app


async def init_test_client() -> AsyncClient:
    async with LifespanManager(app):
        async with AsyncClient(app=app, base_url="http://localhost") as ac:
            yield ac



import os

from zrb.helper.file.text import read_text_file_async, write_text_file_async
from zrb.helper.typecheck import typechecked
from zrb.helper.util import to_kebab_case, to_snake_case
from zrb.task.task import Task


@typechecked
async def create_app_config(
    task: Task, project_dir: str, app_name: str, module_name: str
):
    kebab_app_name = to_kebab_case(app_name)
    snake_module_name = to_snake_case(module_name)
    upper_snake_module_name = snake_module_name.upper()
    config_file_path = os.path.join(
        project_dir, "src", kebab_app_name, "src", "config.py"
    )
    config_code = "\n".join(
        [
            f"APP_ENABLE_{upper_snake_module_name}_MODULE = str_to_boolean(os.environ.get(",  # noqa
            f"    'APP_ENABLE_{upper_snake_module_name}_MODULE', 'true'" "))",
        ]
    )
    task.print_out(f"Read config from: {config_file_path}")
    code = await read_text_file_async(config_file_path)
    code += "\n" + config_code
    task.print_out(f"Write config to: {config_file_path}")
    await write_text_file_async(config_file_path, code)



from zrb.builtin.project.add._group import project_add_group
from zrb.task_group.group import Group

project_add_task_group = Group(
    name="task", description="Add task to project", parent=project_add_group
)



from zrb.helper.default_env import inject_default_env
from zrb.runner import runner
from zrb.task.any_task import AnyTask
from zrb.task.any_task_event_handler import (
    OnFailed,
    OnReady,
    OnRetry,
    OnSkipped,
    OnStarted,
    OnTriggered,
    OnWaiting,
)
from zrb.task.base_remote_cmd_task import BaseRemoteCmdTask, RemoteConfig
from zrb.task.checker import Checker
from zrb.task.cmd_task import CmdTask
from zrb.task.decorator import python_task
from zrb.task.docker_compose_start_task import DockerComposeStartTask
from zrb.task.docker_compose_task import DockerComposeTask, ServiceConfig
from zrb.task.flow_task import FlowTask
from zrb.task.http_checker import HTTPChecker
from zrb.task.notifier import Notifier
from zrb.task.parallel import AnyParallel, Parallel
from zrb.task.path_checker import PathChecker
from zrb.task.path_watcher import PathWatcher
from zrb.task.port_checker import PortChecker
from zrb.task.recurring_task import RecurringTask
from zrb.task.remote_cmd_task import RemoteCmdTask
from zrb.task.resource_maker import (
    Replacement,
    ReplacementMutator,
    ResourceMaker,
    get_default_resource_excludes,
    get_default_resource_skip_parsing,
)
from zrb.task.rsync_task import RsyncTask
from zrb.task.server import Controller, Server
from zrb.task.task import Task
from zrb.task.time_watcher import TimeWatcher
from zrb.task.watcher import Watcher
from zrb.task.wiki_task import create_wiki_tasks
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_group.group import Group
from zrb.task_input.any_input import AnyInput
from zrb.task_input.bool_input import BoolInput
from zrb.task_input.choice_input import ChoiceInput
from zrb.task_input.float_input import FloatInput
from zrb.task_input.int_input import IntInput
from zrb.task_input.multiline_input import MultilineInput
from zrb.task_input.password_input import PasswordInput
from zrb.task_input.str_input import StrInput
from zrb.task_input.task_input import Input

assert create_wiki_tasks
assert runner
assert AnyTask
assert OnTriggered
assert OnWaiting
assert OnSkipped
assert OnStarted
assert OnReady
assert OnRetry
assert OnFailed
assert AnyParallel
assert Parallel
assert python_task
assert Task
assert CmdTask
assert DockerComposeTask
assert DockerComposeStartTask
assert ServiceConfig
assert BaseRemoteCmdTask
assert RemoteConfig
assert RemoteCmdTask
assert RsyncTask
assert Notifier
assert Checker
assert HTTPChecker
assert PortChecker
assert PathChecker
assert PathWatcher
assert Controller
assert Server
assert Watcher
assert TimeWatcher
assert ResourceMaker
assert FlowTask
assert RecurringTask
assert get_default_resource_excludes
assert get_default_resource_skip_parsing
assert Replacement
assert ReplacementMutator
assert AnyInput
assert Input
assert BoolInput
assert ChoiceInput
assert FloatInput
assert IntInput
assert MultilineInput
assert PasswordInput
assert StrInput
assert Env
assert EnvFile
assert Group

inject_default_env()



from zrb.helper.accessories.name import get_random_name
from zrb.task_input.int_input import IntInput
from zrb.task_input.str_input import StrInput

package_name_input = StrInput(
    name="package-name",
    description="Package name",
    prompt="Package name",
    default=get_random_name(),
)

generator_name_input = StrInput(
    name="generator-name",
    description="Generator name",
    prompt="Generator name",
    default=get_random_name(),
)

generator_base_image_input = StrInput(
    name="generator-base-image",
    description="Base image",
    prompt="Base image",
    default="python:3.10-slim",
)

generator_app_port_input = IntInput(
    name="generator-app-port",
    description="Default app port",
    prompt="Default app port",
    default="8080",
)



from collections.abc import Mapping
from logging import Logger
from typing import Any

from component.messagebus import Publisher
from component.repo import SearchFilter
from component.rpc import Caller, Server
from module.auth.schema.token import AccessTokenData
from module.log.integration.model.activity_model import activity_model
from module.log.schema.activity import ActivityData


def register_rpc(
    logger: Logger, rpc_server: Server, rpc_caller: Caller, publisher: Publisher
):
    logger.info('🥪 Register RPC handlers for "log.activity"')

    @rpc_server.register("log_get_activity")
    async def get(
        keyword: str,
        criterion: Mapping[str, Any],
        limit: int,
        offset: int,
        user_token_data: Mapping[str, Any],
    ) -> Mapping[str, Any]:
        result = await activity_model.get(
            search_filter=SearchFilter(keyword=keyword, criterion=criterion),
            limit=limit,
            offset=offset,
        )
        return result.model_dump()

    @rpc_server.register("log_get_activity_by_id")
    async def get_by_id(
        id: str, user_token_data: Mapping[str, Any] = {}
    ) -> Mapping[str, Any]:
        row = await activity_model.get_by_id(id)
        return row.model_dump()

    @rpc_server.register("log_insert_activity")
    async def insert(
        data: Mapping[str, Any], user_token_data: Mapping[str, Any]
    ) -> Mapping[str, Any]:
        user_token_data = AccessTokenData(**user_token_data)
        data["created_by"] = user_token_data.user_id
        data["updated_by"] = user_token_data.user_id
        row = await activity_model.insert(data=ActivityData(**data))
        return row.model_dump()

    @rpc_server.register("log_update_activity")
    async def update(
        id: str, data: Mapping[str, Any], user_token_data: Mapping[str, Any]
    ) -> Mapping[str, Any]:
        user_token_data = AccessTokenData(**user_token_data)
        data["updated_by"] = user_token_data.user_id
        row = await activity_model.update(id=id, data=ActivityData(**data))
        return row.model_dump()

    @rpc_server.register("log_delete_activity")
    async def delete(id: str, user_token_data: Mapping[str, Any]) -> Mapping[str, Any]:
        user_token_data = AccessTokenData(**user_token_data)
        row = await activity_model.delete(id=id)
        return row.model_dump()



import asyncio
import os
from collections.abc import Mapping
from typing import Any

import jsons
from dotenv import dotenv_values

from zrb.helper.docker_compose.file import add_services
from zrb.helper.file.text import (
    append_text_file_async,
    read_text_file_async,
    write_text_file_async,
)
from zrb.helper.typecheck import typechecked
from zrb.helper.util import to_kebab_case, to_snake_case
from zrb.task.task import Task


@typechecked
async def create_microservice_config(
    task: Task, project_dir: str, app_name: str, module_name: str
):
    modules = await _create_automation_json_config(
        task, project_dir, app_name, module_name
    )
    return await asyncio.gather(
        asyncio.create_task(
            _add_docker_compose_service(
                task, modules, project_dir, app_name, module_name
            )
        ),
        asyncio.create_task(
            _append_compose_env(task, modules, project_dir, app_name, module_name)
        ),
    )


@typechecked
async def _add_docker_compose_service(
    task: Task, modules: list[str], project_dir: str, app_name: str, module_name: str
):
    kebab_app_name = to_kebab_case(app_name)
    snake_module_name = to_snake_case(module_name)
    upper_snake_module_name = snake_module_name.upper()
    module_app_port = 8080 + len(modules)
    module_app_port_str = str(module_app_port)
    docker_compose_file_path = os.path.join(
        project_dir, "src", kebab_app_name, "docker-compose.yml"
    )
    app_container_port_env_name = f"APP_{upper_snake_module_name}_MODULE_PORT"
    app_container_port_env = (
        "${" + app_container_port_env_name + ":-" + module_app_port_str + "}"
    )
    app_host_port_env_name = f"APP_{upper_snake_module_name}_HOST_MODULE_PORT"
    app_host_port_env = "${" + app_host_port_env_name + ":-" + module_app_port_str + "}"
    service_definition = _get_new_docker_compose_service_definition(
        app_name=app_name,
        module_name=module_name,
        app_host_port_env=app_host_port_env,
        app_container_port_env=app_container_port_env,
    )
    task.print_out(f"Add service at: {docker_compose_file_path}")
    add_services(
        file_name=docker_compose_file_path,
        new_services=service_definition,
    )


@typechecked
def _get_new_docker_compose_service_definition(
    app_name: str, module_name: str, app_host_port_env: str, app_container_port_env: str
) -> Mapping[str, Any]:
    kebab_app_name = to_kebab_case(app_name)
    kebab_module_name = to_kebab_case(module_name)
    snake_module_name = to_snake_case(module_name)
    upper_snake_module_name = snake_module_name.upper()
    return {
        f"{kebab_app_name}-{kebab_module_name}-service": {
            "build": {"dockerfile": "Dockerfile", "context": "./src"},
            "image": "${IMAGE:-" + kebab_app_name + "}",
            "container_name": "${CONTAINER_PREFIX:-my}-"
            + f"{kebab_app_name}-{kebab_module_name}-service",  # noqa
            "hostname": f"{kebab_app_name}-{kebab_module_name}-service",
            "env_file": ["src/template.env", "all-module-disabled.env"],
            "environment": {
                "APP_NAME": "${APP_NAME:-"
                + kebab_app_name
                + "}-"
                + f"{kebab_module_name}-service",  # noqa
                "APP_ENABLE_OTEL": "${APP_ENABLE_OTEL:-0}",
                "APP_PORT": app_container_port_env,
                "APP_ENABLE_EVENT_HANDLER": "true",
                "APP_ENABLE_RPC_SERVER": "true",
                "APP_ENABLE_API": "false",
                "APP_ENABLE_FRONTEND": "false",
                f"APP_ENABLE_{upper_snake_module_name}_MODULE": "true",
            },
            "ports": [f"{app_host_port_env}:{app_container_port_env}"],
            "restart": "unless-stopped",
            "profiles": ["microservices"],
            "healthcheck": {
                "test": f"curl --fail http://localhost:{app_container_port_env}/readiness || killall uvicorn",  # noqa
                "interval": "20s",
                "timeout": "3s",
                "retries": 10,
                "start_period": "20s",
            },
            "networks": ["zrb"],
        }
    }


@typechecked
async def _create_automation_json_config(
    task: Task,
    project_dir: str,
    app_name: str,
    module_name: str,
):
    snake_app_name = to_snake_case(app_name)
    snake_module_name = to_snake_case(module_name)
    json_modules_file_path = os.path.join(
        project_dir, "_automate", snake_app_name, "config", "modules.json"
    )
    task.print_out(f"Read json config from: {json_modules_file_path}")
    json_str = await read_text_file_async(json_modules_file_path)
    task.print_out(f'Add "{snake_module_name}" to json config')
    modules: list[str] = jsons.loads(json_str)
    modules.append(snake_module_name)
    json_str = jsons.dumps(modules)
    task.print_out(f"Write new json config to: {json_modules_file_path}")
    await write_text_file_async(json_modules_file_path, json_str)
    return modules


@typechecked
async def _append_compose_env(
    task: Task, modules: list[str], project_dir: str, app_name: str, module_name: str
):
    kebab_app_name = to_kebab_case(app_name)
    snake_module_name = to_snake_case(module_name)
    upper_snake_module_name = snake_module_name.upper()
    compose_template_env_path = os.path.join(
        project_dir, "src", kebab_app_name, "docker-compose.env"
    )
    compose_env_map = dotenv_values(compose_template_env_path)
    host_port = int(compose_env_map.get("APP_GATEWAY_HOST_PORT", "8080"))
    module_app_port = host_port + len(modules)
    new_env_str = "\n".join(
        [
            f"APP_{upper_snake_module_name}_HOST_MODULE_PORT={module_app_port}",
            f"APP_{upper_snake_module_name}_MODULE_PORT={module_app_port}",
        ]
    )
    task.print_out(f"Add new environment to: {compose_template_env_path}")
    await append_text_file_async(compose_template_env_path, new_env_str)



from zrb import Group
from zrb.builtin import project_group

snake_zrb_app_name_group = Group(
    name="kebab-zrb-app-name",
    parent=project_group,
    description="Manage human readable zrb app name",
)



from zrb.builtin.ubuntu._group import ubuntu_group
from zrb.builtin.ubuntu.install import (
    install_ubuntu_all,
    install_ubuntu_essentials,
    install_ubuntu_tex,
    install_ubuntu_toys,
    ubuntu_install_group,
)
from zrb.builtin.ubuntu.update import update_ubuntu

assert ubuntu_group
assert ubuntu_install_group
assert update_ubuntu
assert ubuntu_install_group
assert install_ubuntu_all
assert install_ubuntu_essentials
assert install_ubuntu_tex
assert install_ubuntu_toys



import os

from zrb.builtin.devtool.install._group import devtool_install_group
from zrb.builtin.devtool.install._helper import write_config
from zrb.runner import runner
from zrb.task.cmd_task import CmdTask
from zrb.task.flow_task import FlowTask
from zrb.task.task import Task
from zrb.task_input.str_input import StrInput

_CURRENT_DIR = os.path.dirname(__file__)
_SHELL_SCRIPT_DIR = os.path.join(_CURRENT_DIR, "..", "..", "..", "..", "shell-scripts")

install_tmux = FlowTask(
    name="tmux",
    group=devtool_install_group,
    description="Terminal multiplexer",
    inputs=[
        StrInput(
            name="tmux-config-file",
            shortcut="c",
            prompt="Tmux config file",
            default="~/.tmux.conf",
        )
    ],
    steps=[
        CmdTask(
            name="install-tmux",
            cmd_path=[
                os.path.join(_SHELL_SCRIPT_DIR, "_common-util.sh"),
                os.path.join(_CURRENT_DIR, "install.sh"),
            ],
            preexec_fn=None,
        ),
        Task(
            name="configure-tmux",
            run=write_config(
                template_file=os.path.join(_CURRENT_DIR, "resource", "config.sh"),
                config_file="{{input.tmux_config_file}}",
            ),
        ),
    ],
    retry=0,
)
runner.register(install_tmux)



from .inject_delete_page import inject_delete_page
from .inject_detail_page import inject_detail_page
from .inject_insert_page import inject_insert_page
from .inject_list_page import inject_list_page
from .inject_repo import inject_repo
from .inject_schema import inject_schema
from .inject_test import inject_test
from .inject_update_page import inject_update_page

assert inject_list_page
assert inject_detail_page
assert inject_insert_page
assert inject_update_page
assert inject_delete_page
assert inject_repo
assert inject_schema
assert inject_test



import os

from zrb import CmdTask, runner

from ..._project import publish_project
from .._group import snake_zrb_package_name_group
from ..build import build_snake_zrb_package_name

_CURRENT_DIR = os.path.dirname(__file__)
_PKG_AUTOMATE_DIR = os.path.dirname(_CURRENT_DIR)
_PROJECT_DIR = os.path.dirname(os.path.dirname(_PKG_AUTOMATE_DIR))
_PKG_RESOURCE_DIR = os.path.join(_PROJECT_DIR, "src", "kebab-zrb-package-name")

publish_snake_zrb_package_name = CmdTask(
    name="publish",
    description="Publish human readable zrb package name",
    group=snake_zrb_package_name_group,
    upstreams=[build_snake_zrb_package_name],
    cwd=_PKG_RESOURCE_DIR,
    cmd_path=[
        os.path.join(_PKG_AUTOMATE_DIR, "activate-venv.sh"),
        os.path.join(_CURRENT_DIR, "publish.sh"),
    ],
    should_print_cmd_result=False,
)

publish_snake_zrb_package_name >> publish_project

runner.register(publish_snake_zrb_package_name)
